# sb::Driver::SqlBase
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==

# ======================================
package sb::Driver::SqlBase;
use strict;
use DBI;
use sb::Driver ();
use base qw( sb::Driver );
use vars qw( $VERSION );
$VERSION = '0.14';
# 0.14 [2007/12/04] defined insert as overridden method / ignore some dying points
# 0.13 [2007/09/24] fixed a simple syntax bug in generate_where
# 0.12 [2007/09/14] added new parameter for table name prefix
# 0.11 [2007/08/12] fixed a bug matched_number returns the wrong number
# 0.10 [2007/08/07] changed load and load_list to add "limit" in %param
# 0.09 [2007/08/04] changed check_order to correct "ORDER BY" statement
# 0.08 [2007/08/03] added insert to convert old data
# 0.07 [2007/07/20] fixed a bug in save_list
# 0.06 [2007/07/16] changed return to make them consistent
# 0.05 [2007/02/14] changed save_list
# 0.04 [2006/07/04] got rid of "return( undef )"
# 0.03 [2006/03/29] added count adn count_list
# 0.02 [2006/03/17] changed bracket rule
# 0.01 [2006/03/06] changed generate_where to use SQL92 syntax
# 0.00 [2006/03/04] generated
# --------------------------------------
# // defines / static variables
sub DEFAULT_HEAD (){ 'sb_' };
# --------------------------------------
# // override base class
sub setup
{
	my $self = shift;
	return $self->SUPER::setup(
		'dbh'        => undef, # database handler
		'sql_buffer' => undef, # sql buffer for where phrase
		'sql_place'  => undef, # sql buffer for place array
		'prefix'     => undef, # table name prefix
		@_
	);
}
sub finish
{
	my $self = shift;
	$self->dbh->disconnect if ($self->dbh);
	return;
}
# --------------------------------------
# // override base class - instance method
sub name
{
	my $self = shift;
	my ($class,$elem) = @_;
	return $self->SUPER::name($class) . '_' . $elem;
}
sub matched_number
{
	my $self = shift;
	return $self->{'num'} if (! $self->{'sql_buffer'} or ! $self->{'sql_place'});
	$self->{'num'} = 0; # reset count
	my $sql = 'SELECT COUNT(*)' . $self->{'sql_buffer'};
	my $cnt = 0;
	# get statement handler
	my $sth = $self->dbh->prepare($sql) or return; # or die($self->dbh->errstr);
	return if (!$sth);
	$sth->execute(@{$self->{'sql_place'}}) or return; # or die($self->dbh->errstr);
	$sth->bind_columns(undef,\$cnt); # the first "undef" is for backward compatibility
	$sth->fetch or return;
	$sth->finish;
	# set member variables
	$self->{'sql_buffer'} = undef;
	$self->{'sql_place'} = undef;
	$self->{'num'} = $cnt;
	return $self->{'num'};
}
# --------------------------------------
# // override base class - implementation
sub load($$;@)
{
	my $self  = shift;
	my $class = shift; # get sub-class name of sb::Data::Object
	my %param = ( # input paramaeters
		'cond'  => {},    # searching option
		'id'    => undef, # inidicates id to load directly
		'logic' => 'and', # logic factor for searching option
		'bgn'   => 0,     # offset of objects to load
		'num'   => -1,    # maximum of objects to load
		'sort'  => 'id',  # sorting option
		'order' => 0,     # sorting order 0:descent 1:ascent
		'limit' => undef, # parameter limitation
		@_
	);
	return if ( !$self->dbh );
	return $self->load_list($class,%param);
}
sub count($$;@)
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my %param = ( # input paramaeters
		'cond'  => {},    # searching option
		'id'    => undef, # inidicates id to load directly
		'logic' => 'and', # logic factor for searching option
		'bgn'   => 0,     # offset of objects to load
		'num'   => -1,    # maximum of objects to load
		'sort'  => 'id',  # sorting option
		'order' => 0,     # sorting order 0:descent 1:ascent
		@_
	);
	return if ( !$self->dbh );
	return $self->count_list($class,%param);
}
sub save($@)
{
	my $self = shift;
	my @objs = @_; # set instances of sb::Data::Object to save
	return if ( $#objs == -1 );
	return if ( !$self->dbh );
	my $class = ref($objs[0]);
	return $self->save_list($class,@objs);
}
sub insert($$)
{
	my $self = shift;
	my $obj = shift; # an instance of sb::Data::Object
	my $class = ref($obj);
	return if ( !$self->dbh or !$obj or !$class );
	my @sets = $class->elements;
	my @field = map { $self->name($class,$_) } @sets;
	my $sql = 'INSERT INTO ' .  $self->table_name($class);
	$sql .= ' (' . join(', ',@field) . ') VALUES (' . join(', ',('?') x @field) . ')';
	my $sth = $self->dbh->prepare($sql);
	return if (!$sth);
	for (my $i=0;$i<@sets;$i++)
	{
		my $elem = _linefeed($obj->{$sets[$i]});
		$sth->bind_param($i+1,$elem);
	}
	$sth->execute() or return;
	$sth->finish;
	return 1;
}
sub new_id($$)
{
	my $self = shift;
	my $class = shift; # sets class name of sb::Data::Object
	return if ( !$self->dbh );
	my $id = $self->prepare_id;
	my @target = ();
	my @values = ();
	foreach my $elem ( $class->elements )
	{
		next if ($elem eq 'id' and ! defined($id));
		push(@target,$self->name($class,$elem));
		push(@values,($elem ne 'id') ? 'NULL' : $id);
	}
	my $sql = 'INSERT INTO ' .  $self->table_name($class);
	$sql .= ' (' . join(', ',@target) . ') ';
	$sql .= 'VALUES (' . join(', ', @values) . ')';
	my $sth = $self->dbh->prepare($sql);
	return if (!$sth);
	$sth->execute() or return; # or die($self->dbh->errstr);
	$sth->finish;
	$id = $self->fetch_id($sth) if (! defined($id));
	return $id;
}
sub create_table($$)
{
	my $self = shift;
	my $class = shift; # set class name of sb::Data::Object
	return if ( !sb::load_module($class) );
	return if ( !$self->dbh );
	my @sql = ();
	my %table = $class->table;
	foreach my $elem ( $class->elements )
	{
		my $type = $self->convert_type($table{$elem});
		push(@sql, $self->name($class,$elem) . ' ' . $type);
	}
	my $state = 'CREATE TABLE ' . $self->table_name($class);
	$state .= ' (' . join(', ',@sql) . ')';
	$self->dbh->do($state) or return;
	return 1;
}
# --------------------------------------
# // public functions - declaration
sub prepare_id
{ # prepare id before inserting
	my $self = shift;
	return;
}
sub fetch_id
{ # fetch id after insering
	my $self = shift;
	return;
}
# --------------------------------------
# // public functions - accessor
sub table_name
{
	my $self = shift;
	my $name = $self->SUPER::name(shift);
	return ($self->{'prefix'})
		? $self->{'prefix'} . '_' . $name
		: DEFAULT_HEAD . $name;
}
sub dbh
{
	my $self = shift;
	$self->{'dbh'} = shift if (@_);
	return $self->{'dbh'};
}
sub count_list
{
	my $self  = shift;
	my $class = shift; # set class name of sb::Data::Object
	my %param = @_;
	# initialize local variables
	my $label = $self->name($class,'');
	my $logic = $self->check_logic($param{'logic'});
	my ($where,$place) = $self->generate_where($label,$param{'id'},$param{'cond'});
	# buffer sql statement
	$self->{'sql_buffer'}  = ' FROM ' . $self->table_name($class);
	$self->{'sql_buffer'} .= ' WHERE ' . join($logic,@{$where}) if (@{$where});
	$self->{'sql_place'} = $place;
	return $self->matched_number();
}
sub load_list
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my %param = @_;
	my @array = ();
	# initialize local variables
	$param{'num'} = int($param{'num'});
	$param{'bgn'} = ($param{'num'} <= 0) ? 0 : int($param{'bgn'});
	my $sth = undef; # statment handler
	my $label = $self->name($class,'');
	my $logic = $self->check_logic($param{'logic'});
	my $order = $self->check_order($label,$param{'sort'},$param{'order'});
	my ($where,$place) = $self->generate_where($label,$param{'id'},$param{'cond'});
	my $sql = '';
	# prepare binds
	my (%dat,@binds);
	my @elements = ($param{'limit'}) ? $self->param_list($class,$param{'limit'}) : $class->elements;
	foreach my $val ( @elements )
	{
		push(@binds,\$dat{$val});
	}
	# buffer sql statement
	$self->{'sql_buffer'}  = ' FROM ' . $self->table_name($class);
	$self->{'sql_buffer'} .= ' WHERE ' . join($logic,@{$where}) if (@{$where});
	$self->{'sql_place'} = $place;
	# generate sql statement
	$sql .= 'SELECT ';
	$sql .= ($param{'limit'}) ? join(',',map { $self->name($class,$_) } @elements) : '*';
	$sql .= $self->{'sql_buffer'};
	$sql .= ' ORDER BY ' . $order;
	if ($param{'num'} >= 0)
	{
		$sql .= ' LIMIT ' . $param{'num'};
		$sql .= ' OFFSET ' . $param{'bgn'};
	}
	# get statement handler
	$sth = $self->dbh->prepare($sql) or return; # or die($self->dbh->errstr);
	return if (!$sth);
	$sth->execute(@{$place}) or return; # or die($self->dbh->errstr);
	$sth->bind_columns(undef,@binds); # the first "undef" is for backward compatibility
	while ($sth->fetch)
	{
		push(@array,$class->create(%dat));
	}
	$sth->finish;
	return @array;
}
sub save_list
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my @objs  = @_;
	my $sth = undef; # statment handler
	my $table = $self->table_name($class);
	my @sets = grep { $_ ne 'id' } $objs[0]->elements;
	my $set  = join(', ',map { $self->name($class,$_) . '=?' } @sets);
	my $whph = ' WHERE ' . $self->name($class,'id') . '=';
	foreach my $obj ( @objs )
	{
		next if ( ref($obj) ne $class ); # ignore different object
		my $sql = '';
		my $where = $whph . $obj->id;
		if ($obj->erased)
		{
			$sql = 'DELETE FROM ' . $table . $where;
		}
		else
		{
			$sql = 'UPDATE ' . $table . ' SET ' . $set . $where;
		}
		$sth = $self->dbh->prepare($sql) or return; # or die($self->dbh->errstr);
		if (!$obj->erased)
		{
			for (my $i=0;$i<@sets;$i++)
			{
				my $elem = _linefeed($obj->{$sets[$i]});
				$sth->bind_param($i+1,$elem);
			}
		}
		$sth->execute or return; # or die($self->dbh->errstr);
		$sth->finish;
	}
	return 1;
}
# --------------------------------------
# // public functions - sql tools
sub quote
{
	my $self = shift;
	return $self->dbh->quote(@_);
}
sub generate_where
{
	my $self = shift;
	my ($name,$id,$cond) = @_;
	my (@where,@place);
	if ($id ne '')
	{
		push(@where,$name . 'id=?');
		push(@place,int($id));
	}
	else
	{
		# [search string condition] ===
		my @like = ();
		foreach my $key ( keys(%{$cond}) )
		{
			next if ( $key eq '__range' or $key eq '__combo' );
			if ( $cond->{$key} =~ /^\/(.*)\/$/ )
			{
				push(@like,$name . $key . ' LIKE ?'); # almost same as /\Q$word\E/i
				push(@place,'%' . $1 . '%');
				$cond->{$key} = undef; # clear condition
			}
		}
		push(@where,'(' . join(' OR ',@like) . ')') if (@like);
		# [other conditions] ===
		foreach my $key ( keys(%{$cond}) )
		{
			next if ( $key eq '__range' or $key eq '__combo' );
			next if (! defined($cond->{$key})); # skip if $cond->{$key} is undef
			if ( ref($cond->{$key}) eq 'ARRAY' )
			{ # array or range
				my $phrase = '(';
				my $label = $name . $key;
				if (defined($cond->{'__range'}->{$key}))
				{
					# range ( between the first value and the second value )
					if ( $cond->{'__range'}->{$key} eq 'zone' )
					{
						# timezone range, zone value should be offset of value
						$phrase .= $label . '-' . $name . 'zone>=? AND ' . $label . '-' . $name . 'zone<?';
					}
					else
					{
						$phrase .= $label . '>=? AND ' . $label . '<?';
					}
					push(@place,$cond->{$key}[0],$cond->{$key}[1]);
				}
				else
				{
					# array ( check matching one of list values )
					$phrase .= join(' OR ',map { $label . '=?' } @{$cond->{$key}});
					push(@place,@{$cond->{$key}});
				}
				$phrase .= ')';
				push(@where,$phrase);
			}
			else
			{ # normal or combo
				my $phrase = $name . $key . '=?';
				push(@place,$cond->{$key});
				if (defined($cond->{'__combo'}->{$key}))
				{
					my %combo = %{$cond->{'__combo'}};
					$phrase = '(' . $phrase . ' OR ';
					$phrase .= $name . $combo{$key} . ' LIKE ?';
					push(@place,'%' . $combo{$combo{$key}} . '%');
					$phrase .= ')'; # we should close phrase.
				}
				push(@where,$phrase);
			}
		} # end of [other conditions]
	}
	return (\@where,\@place);
}
sub check_logic
{
	my $self = shift;
	my $logic = shift;
	return ' OR ' if ($logic eq 'or');
	return ' AND ';
}
sub check_order
{
	my $self = shift;
	my $label = shift;
	my $check = shift;
	my $order = (' ASC',' DESC')[shift];
	my $out = $label . 'id' . $order;
	$out = $label . $check . $order . ',' . $out if ($check ne 'id');
	return $out;
}
# --------------------------------------
# // private functions - tools
sub _linefeed
{ # unify linefeed code
	$_[0] =~ s/\x0D\x0A/\n/g;
	$_[0] =~ tr/\x0D\x0A/\n\n/;
	while ($_[0] =~ /\n$/)
	{
		$_[0] =~ s/\n$//g;
	}
	return $_[0];
}
1;
__END__
